c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: driver test programm for i-smooth.

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      parameter ( maxint = 102400 )

      dimension rate(100), ratenew(maxint), frate(4)

      dimension time(0:100), timenew(0:maxint)

      print *,' enter k, the number of rates'

      read *, k

      print *,' enter the rate values'

      read *, (rate(i), i=1,k)

      print *,' enter context: 1 for finite horizon, 0 for cyclic'

      read *, ifinite

      frate(1) = -1d0

      frate(2) = -1d0

      frate(3) = -1d0

      frate(4) = -1d0    

      if (ifinite .eq. 1) then

         print *, 'how many terminal fixed rates on the left and right'

         read *, nl, nr

      endif

      if (nl .eq. 1) then

         print *,' enter the rate on the left'

         read *, frate(2)

      elseif (nl .eq. 2) then

         print *,' enter the two rates on the left'

         read *, frate(1), frate(2)

      endif

      if (nr .eq. 1) then

         print *,' enter the rate on the right'

         read *, frate(3)

      elseif (nr .eq. 2) then

         print *,' enter the two rates on the right'

         read *, frate(3), frate(4)

      endif

c

      print *,' enter time(0) and the end time for each interval'

      read *, (time(i),i=0,k)

c

 10   print *, ' enter number of ismooth iterations'

      read (*,*, end=30) niter

      k2pn = k * 2**niter

      if (k2pn .gt. maxint) then

         print *,' Error:  The arrays ratenew and timenew are'

         print *,'    dimensioned for no more than ten iterations'

         print *,' '

         go to 10

      endif

c

c     ...compute the smoothed rates and associated times

      call ismooth(k,time,rate,ifinite,frate,niter,timenew,ratenew,obj)

c

c

      print *,' '

      print *,' ..... I-SMOOTH Results .....'

      print *,' # of rates, k =', k

      print *,' rates =', (rate(i), i=1,k)

      if (ifinite .eq. 0) then

         print *,' cyclic context'        

      else

         print *,' finite-horizon context'

         if (nl .gt. 0) then

           print *,' left-side fixed rates:', frate(1), frate(2)

         endif

         if (nr .gt. 0) then

           print *,' right-side fixed rates:', frate(3), frate(4)

         endif

      endif

      print *,' objective-function value =', obj

      k2pn = k * 2**niter

      print *,' number of new intervals =', k2pn

      print *,' new intervals: number, begin time, end time, rate...'

      print *,' '

      do i = 1, k2pn

         print *, i, timenew(i-1), timenew(i), ratenew(i)

      enddo

      print *,' '

      go to 10

 30   stop 999

      end







      subroutine ismooth( k0, time0, rate0, ifinite0, frate0, niter,

     &                                             time, rate, obj )

c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: execute niter iterations of the i-smooth algorithm.

c     input:  k0       = number of initial time intervals

c             time0    = vector of initial times

c             rate0    = vector of initial rates

c             ifinite0 = 0 if cyclic, 1 if finite-horizon context

c             frate0   = vector of four initial terminal fixed rates

c             niter    = number of i-smooth iterations

c     output: time     = vector of (k * 2^niter) new times

c             rate     = vector of (k * 2^niter) new rates

c             obj      = objective function (from the last iteration)

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension  time0(0:1),rate0(1),frate0(4),time(0:1),rate(1)

      dimension  frate(4), gamma(k0 * 2**niter)

      print *,' ...beginning ismooth'

      k = k0

      ifinite = ifinite0

      time(0) = time0(0)

      do i = 1, k

         rate(i) = rate0(i)

         time(i) = time0(i)

      enddo

      do i = 1, 4

         frate(i) = frate0(i)

      enddo

c

c     ...i-smooth's main loop

      do j = 1, niter

         print *,' '

         print *,'    ISMOOTH: ITERATION =', j

c

c        ...smooth the rates by doubling the number of intervals

         call smooth( k, rate, ifinite, frate, gamma, obj )

c

c        ...compute the 2k new rates and times

         do i = 1, k

            rate(2*(k-i)+2) = rate(k-i+1) + gamma(k-i+1)

            rate(2*(k-i)+1) = rate(k-i+1) - gamma(k-i+1)

            time(2*(k-i)+2) = time(k-i+1) 

            time(2*(k-i)+1) = ( time(k-i+1) + time(k-i) ) / 2d0

         enddo

c

c        ...update the terminal fixed rates on both ends

         if ((ifinite .eq. 1) .and. (j .eq. 1)) then

            frate(1) = frate(2)

            frate(4) = frate(3)

         endif

         k = 2*k

c

         do i = 1, k

            print *, i, time(i-1), time(i), rate(i)

         enddo

      enddo

      end







      subroutine smooth( k0, rate0, ifinite0, frate0, gamma0, obj )

c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: one iteration of the i-smooth algorithm.

c     input:  k0       = number of initial time intervals

c             rate0    = vector of initial rates

c             ifinite0 = 0 if cyclic, 1 if finite-horizon context

c             frate0   = vector of four initial terminal fixed rates

c     output: gamma0   = vector of k0 rates increments

c             obj      = objective-function value

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension  rate0(1), frate0(4), gamma0(1)

      dimension  rate(k0), frate(4), gamma(k0), ilocked(k0), isolved(k0)

      print *,' ...beginning smooth'

c     ...initialize the ilocked vector: 1 implies locked, else 0

      nlocked = 0

      do i = 1, k0

         ilocked(i) = 0

         if (rate0(i) .le. 0d0) then

            gamma0(i)  = 0d0

            ilocked(i) = 1

            nlocked    = nlocked + 1

         endif

      enddo

c

 10   continue

c        ...outer loop.  constrained optimization.

         do i = 1, k0

            isolved(i) = 0

            if (ilocked(i) .eq. 1) isolved(i) = 1

         enddo

         nsolved = nlocked

c

         do while (nsolved .lt. k0) 

c            ...determine ibegin and iend

            call scan( k0, ifinite0, isolved, nsolved, ibegin, iend )

c

c           ...set up the problem from ibegin to iend

            call problem( k0, ifinite0, rate0, frate0, isolved, gamma0, 

     &                   ibegin, iend, k, ifinite, rate, frate, nl, nr )

c

c           ...perform unconstrained (negative is ok) smoothing

            if( ifinite .eq. 0 ) then

               call csmooth( k, rate, gamma )

            else

               call fsmooth( k, rate, frate, nl, nr, gamma )

            endif

c

c           ...set the gamma values and indicate solved

            j = 1

            i = ibegin

 20         gamma0(i)  = gamma(j)      

            isolved(i) = 1         

            nsolved    = nsolved + 1

            if (i .ne. iend) then

               j = j + 1

               i = i + 1

               if (i .gt. k0) i = i - k0

               go to 20

            endif

c            print *,' '

c            print *,' in smooth...'

c            print *,'    k0, ifinite0 =', k0, ifinite0

c            print *,'    frate0  =', (frate0(i),i=1,4)

c            print *,'    rate0   =', (rate0(i),i=1,k0)

c            print *,'    gamma0  =', (gamma0(i),i=1,k0)

c            print *,'    isolved =', (isolved(i),i=1,k0)

c            print *,'    ilocked =', (ilocked(i),i=1,k0)

         enddo

c

c        ...push the most-negative rate to zero.  update gamma and lock.

         call pushup1( k0, rate0, gamma0, ipush )

c         call pushup2( k0, rate0, ifinite0, frate0, gamma0, ipush )

         if (ipush .gt. 0) then

            ilocked(ipush) = 1

            nlocked        = nlocked + 1

c            print *,' '

c            print *,' in smooth, after pushup...'

c            print *,'    gamma0 =', (gamma0(i),i=1,k0)

c            print *,'    isolved =', (isolved(i),i=1,k0)

c            print *,'    ilocked =', (ilocked(i),i=1,k0)

            if (nlocked .lt. k0) go to 10

         endif

      obj = objfunc( k0, rate0, ifinite0, frate0, gamma0 )

      return

      end







      subroutine scan( k, ifinite, isolved, nsolved, ibegin, iend )

c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: scan isolved for the first sequence of unsolved intervals

c     input:   k       = number of intervals

c              ifinite = 0 if cyclic, 1 if finite-horizon context

c              isolved = vector of indicators: 1 if solved, 0 if unsolved

c              nsolved = the number of intervals that are solved: less than k

c     output:  ibegin  = intervals ibegin through interval iend are unsolved

c              iend    = see ibegin

c                        (if cyclic context, then possibly iend < ibegin.)

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension isolved(1)

      print *,'   ...beginning scan'

c     print *,'        k, ifinite, nsolved =', k, ifinite, nsolved

c     print *,'        isolved    =', (isolved(i),i=1,k)

c

      if (nsolved .eq. 0) then

c        ...no intervals are solved

         ibegin = 1

         iend   = k

      elseif (nsolved .eq. k) then

c        ...all intervals are solved.  should not have called "scan".

c         print *,' *****error: nsolved, k =', nsolved, k

         stop 888

      else

c        ...1 .le. nsolved .lt. k.  determine the left-most solved interval

         i = 0

 10      i = i + 1

         if (isolved(i) .eq. 0) go to 10

         isolve1 = i

         nfound  = 1

         iend    = isolve1 - 1

         if (isolve1 .eq. 1) iend = k

         if (nsolved .eq. 1) then

            if (isolve1 .eq. 1) then

               ibegin = 2

            elseif (isolve1 .eq. k) then

               ibegin = 1

            else

               if (ifinite .eq. 0) then

                  ibegin = isolve1 + 1

               else

                  ibegin = 1

               endif

            endif

         else

c           ...1 < nsolved.  repeatedly determine the next solved interval

 20         i = i + 1

            if (isolved(i) .eq. 0) go to 20

            isolve2 = i

            nfound = nfound + 1

            if (isolve2 .eq. isolve1 + 1) then

               isolve1 = isolve2

               if (nfound .lt. nsolved) go to 20

c              ...isolve2 is the right-most solved interval

               if (isolve2 .eq. k) then

                  ibegin = 1

               else

                  ibegin = isolve2 + 1

                  if (ifinite .eq. 1) iend = k

               endif

            else

               ibegin = isolve1 + 1

               iend   = isolve2 - 1

            endif

         endif

      endif

      print *,'          isolved =', (isolved(i),i=1,k)

      print *,'          ibegin, iend =', ibegin, iend

      return

      end







      subroutine problem( k0, ifinite0, rate0, frate0, ilock, gamma0, 

     &                ibegin, iend, k, ifinite, rate, frate, nl, nr )

c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: set up the next problem

c     input:  k0       = number of initial time intervals

c             ifinite0 = 1 if finite horizon, 0 if cyclic

c             rate0    = vector of initial rates

c             frate0   = vector of four initial terminal fixed rates

c             ilock    = vector of locked/unlocked indicators

c             gamma0   = current interval increments

c             ibegin   = first interval of the next problem

c             iend     = last interval of the next problem

c     output: k        = number of intervals in the new problem

c             ifinite  = 1 if finite horizon, 0 if cyclic

c             rate     = vector of rates for the new problem

c             frate    = vector of fixed rates

c             nl       = number of fixed rates on the left side

c             nr       = number of fixed rates on the right side

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension  rate0(1), frate0(4), ilock(1), gamma0(1)

      dimension  rate(1), frate(4)

      print *,'   ...beginning problem'

c

c     ...create the rate vector and determine k

      k = 1

      i = ibegin

 10   rate(k) = rate0(i)

      if (i .ne. iend) then

         i = i + 1

         if (i .gt. k0) i = i - k0

         k = k + 1

         go to 10

      endif

c      print *,'        k, ibegin, iend =', k, ibegin, iend

c     ...determine horizon context.  if cyclic, create problem and return

      if ( (ifinite0 .eq. 0) .and. (k .eq. k0) ) then

         ifinite = 0

         return

      endif

      ifinite = 1

c

c     ...determine fixed rates for the finite-horizon problem

c

c     ...left-side fixed rates

      nl = 0

      if ( (ibegin .eq. 1) .and. (ifinite0 .eq. 1) ) then

         frate(1) = frate0(1)

         frate(2) = frate0(2)

         if (frate(2) .ge. 0d0) nl = 1

         if ( (nl .eq. 1) .and. (frate(1) .ge. 0d0) ) nl = 2

      else

         ibeginm1 = ibegin - 1

         if (ibeginm1 .lt. 1) ibeginm1 = k0

         if (ilock(ibeginm1) .eq. 1) then

            frate(1) = rate0(ibeginm1) - gamma0(ibeginm1)

            frate(2) = rate0(ibeginm1) + gamma0(ibeginm1)

            nl = 2

         endif

      endif

c

c     ...right-side fixed rates

      nr = 0

      if ( (iend .eq. k0) .and. (ifinite0 .eq. 1) ) then

         frate(3) = frate0(3)

         frate(4) = frate0(4)

         if (frate(3) .ge. 0d0) nr = 1

         if ( (nr .eq. 1) .and. (frate(4) .ge. 0d0) ) nr = 2

      else

         iendp1 = iend + 1

         if (iendp1 .gt. k0) iendp1 = 1

         if (ilock(iendp1) .eq. 1) then

            frate(3) = rate0(iendp1) - gamma0(iendp1)

            frate(4) = rate0(iendp1) + gamma0(iendp1)

            nr = 2

         endif

      endif

      return

      end





      subroutine fsmooth( k, rate, frate, nl, nr, gamma )

c     huifen chen and bruce schmeiser, may 15, 2011.

c     purpose: compute the optimal gamma value for the finite-horizon

c              context without considering the nonnegativity constraint.

c     input:  k     = the number of time intervals

c             rate  = vector of initial rates

c             frate = vector of four initial terminal fixed rates

c             nl    = number of fixed rates on the left side

c             nr    = number of fixed rates on the right side

c     output: gamma = the optimal gamma vector

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension rate(1), frate(4), gamma(1)

      dimension p3(0:k+3)

      data kexact/100/      

      data fourth/0.25d0/, half/0.5d0/, threef/0.75d0/

      data two/2d0/, three/3d0/, eight/8d0/

      third = 1d0 / 3d0

      print *,' '

      print *,'   ...beginning fsmooth'

c

c     ...define fixed-rate indicators

      I_L1   = 0

      if (nl .eq. 1) I_L1   = 1

      I_L2   = 0

      if (nl .eq. 2) I_L2   = 1      

      I_Lge1 = 0

      if (nl .ge. 1) I_Lge1 = 1

      I_R1   = 0

      if (nr .eq. 1) I_R1   = 1

      I_R2   = 0

      if (nr .eq. 2) I_R2   = 1      

      I_Rge1 = 0

      if (nr .ge. 1) I_Rge1 = 1

c

      if (k .eq. 1) then

c        ...compute the optimal gamma for k = 1 and return

         if ( (nl .eq. 0) .and. (nr .eq. 0) ) then

            gamma(1) = 0d0

         else

            const = 1d0 / ( 9d0*( I_Lge1 + I_Rge1 ) + ( I_L2 + I_R2 ) )

            gamma(1) = const * ( 

     &                                     I_L2          * frate(1) 

     &       - ( 3 *   I_Lge1 +        2 * I_L2 )        * frate(2) 

     &       + ( 3 * ( I_Lge1 - I_Rge1 ) + I_L2 - I_R2 ) *  rate(1)

     &       + ( 3 *            I_Rge1 +      2 * I_R2 ) * frate(3) 

     &       -                                    I_R2   * frate(4)   )

         endif

         return

      endif

c

c     ...otherwise, k > 1...

c

c     ...compute k_ring: the number of complete time intervals 

c          (including the fixed rates at both ends)

      k_ring = k + I_L2 + I_R2

c

c     ...compute and save powers of three

      p3(0) = 1d0

      do i = 1, k_ring+1

         p3(i) = 3d0 * p3(i-1)

      enddo

c

c     ...compute denominator

         if (k .gt. kexact) then

c           ...for k > kexact, use the large-k approximation

            const = fourth / power( 17, I_L1+I_R1 )

         else

c           ...use the exact computation

            tmp1  = power( 17, I_L1   + I_R1   ) * p3(k_ring - 1)

            tmp2  = power( -1, I_Lge1 + I_Rge1 ) / p3(k_ring - 1)

            const = fourth / ( tmp1 - tmp2 )

         endif

c

c     ...compute the optimal gamma vector.  

      do i = 1, k

         gamma(i) = 0d0

         i_tilde = i + I_L2

         I_one = 0

         if (i .eq. 1) I_one = 1         

         I_k = 0

         if (i .eq. k) I_k = 1            

c

c        ...lambda_{-1} = frate(1), the outer-left fixed rate

c           the weight is zero when nl is not equal to 2.

         if (k .gt. kexact) then

            weight = const * two * I_L2 * power( -1, i-1 ) 

     &                 * power( 17, I_R1 ) / p3(i_tilde)

c            if (i .eq. 1) print *,' -1 approx: k, weight =', k, weight

         else

            if (k_ring - i_tilde - 1 < 0) then

               tmp1 = power( 17, I_R1 ) / p3(i_tilde - k_ring + 1) 

            else

               tmp1 = power( 17, I_R1 ) * p3(k_ring - i_tilde - 1) 

            endif

            tmp2   = power( -1, I_Rge1 ) / p3(k_ring - i_tilde + 1)

            weight = const * two * I_L2 * power( -1, i-1 ) * (tmp1+tmp2)

c            print *,'  -1 exact: weight =', weight

         endif

         gamma(i) = gamma(i) + frate(1) * weight

c           

c        ...lambda_0 = frate(2), the inner-left fixed rate

c           the weight is zero when nl = 0.         

         if (k .gt. kexact) then

            weight = const * power( -1, i ) * power( 17, I_R1 ) 

     &                * (threef * I_Lge1 + half * I_L2) 

     &                * (  power( 17, I_L1 )   * power( 3,  I_L2 ) 

     &                   + power( -1, I_Lge1 ) * power( 3, -I_L2 )  ) 

     &                              / p3(i_tilde - 1)

c            if (i .eq. 1) print *,'  0 approx: k, weight =', k, weight

         else

            weight = const * (threef*I_Lge1 + half*I_L2)*power( -1, i ) 

     &            * (   power( 17, I_L1)   * power( 3,  I_L2 ) 

     &                + power( -1, I_Lge1) * power( 3, -I_L2 )   )

     &            * (   power( 17, I_R1)   * p3(k_ring-i_tilde) 

     &                + power( -1, I_Rge1) / p3(k_ring-i_tilde) )

c               print *,'  0 exact: weight =', weight

         endif

         gamma(i) = gamma(i) + frate(2) * weight

c        

c        ...lambda_1 = rate(1)

         if (k .gt. kexact) then

            tmp1 = threef * I_Lge1 + fourth * I_L2

            weight = const * power( -1, i ) * power( 17, I_R1 )

     &         * ( 1 - ( third + tmp1 ) *I_one )

     &         * (       power( 17, I_L1 ) * power( 3, I_L2 )

     &         * ( 1 - ( three + tmp1 ) * (1-I_one) )

     &          + power( -1, I_Lge1 ) 

     &         * ( 1 - ( third + tmp1 ) * (1-I_one) ) / power( 3, I_L2 ) 

     &                                )    / p3(i_tilde - 1)

c            if (i .eq. 1) print *,'  1 approx: k, weight =', k, weight

         else

            weight = const * power( -1, i ) 

c

     &      * (    power( 17, I_L1 ) * power( 3, I_L2 )

     &        * ( 1 - (three + threef*I_Lge1 + fourth*I_L2) *(1-I_one) )

     &          + power( -1, I_Lge1 ) 

     &        * ( 1 - (third + threef*I_Lge1 + fourth*I_L2) *(1-I_one) )

     &                             / power( 3, I_L2 )               )

c

     &      * (    power( 17, I_R1 ) * p3(k_ring - i_tilde) 

     &        * ( 1 - (third + threef*I_Lge1 + fourth*I_L2) * I_one )

     &           + power( -1, I_Rge1 )

     &        * ( 1 - (three + threef*I_Lge1 + fourth*I_L2) * I_one )  

     &                             / p3(k_ring - i_tilde)          )

c               print *,'  1 exact: weight =', weight

         endif

         gamma(i) = gamma(i) + rate(1) * weight 

c

c        ...lambda_2 to lambda_{i-1}

         do m = 2, i-1

            if (k .gt. kexact) then

               m_tilde = m + I_L2

               weight = const * power(-1, m-i) * eight

     &         * ( power(17, I_R1+ I_L1) / p3(i_tilde - m_tilde + 1) 

     &           - power(17, I_R1)*power(-1, I_Lge1) 

     &                 / (p3(i_tilde) * p3(m_tilde - 1) )

     &           + power(17, I_L1)*power(-1, I_Rge1) 

     &                 / ( p3(k_ring-m_tilde) * p3(k_ring-i_tilde+1) ) )

c               if (i .eq. 1) print *,' m approx:m,k,weight =',m,k,weight

            else

               tmp1 =   power( 17, I_R1 )   * p3(k_ring-i_tilde) 

     &                + power( -1, I_Rge1 ) / p3(k_ring-i_tilde)

               tmp2 =   power( 17, I_L1 )   * p3(m - 2 + I_L2) 

     &                - power( -1, I_Lge1 ) / p3(m     + I_L2)

               weight = const * eight * power( -1, m-i ) * tmp1 * tmp2

c               print *,'  2--i-1 exact: weight =', weight

            endif

            gamma(i) = gamma(i) + rate(m) * weight

         enddo

c                    

         if ( (i .ne. 1) .and. (i .ne. k) ) then

c        ...lambda_i = rate(i), when i is neither 1 nor k

            if (k .gt. kexact) then

               weight = const * eight * 

     &                (    power(17, I_L1) * power( -1, I_Rge1 ) 

     &            / ( p3(k_ring - i_tilde) * p3( k_ring - i_tilde+1)   )

     &              - power( 17, I_R1 ) * power( -1, I_Lge1 ) 

     &              / ( p3(i_tilde) * p3(i_tilde - 1) )  )

c               if (i .eq. 1) print *,'  i approx: k, weight =', k,weight

            else

               if (2*i_tilde - k_ring - 2 > 0) then

                  tmp1 =       p3(2*i_tilde - k_ring - 2)

               else

                  tmp1 = 1d0 / p3(k_ring - 2*i_tilde + 2)

               endif

               if (k_ring - 2*i_tilde > 0) then

                  tmp2 =       p3(k_ring - 2*i_tilde)

               else

                  tmp2 = 1d0 / p3(2*i_tilde - k_ring)

               endif

               weight = const * eight

     &             * (  power( 17, I_L1 ) * power( -1, I_Rge1 ) * tmp1 

     &                - power( 17, I_R1 ) * power( -1, I_Lge1 ) * tmp2 )

c               print *,'  i exact: weight =', weight

            endif

            gamma(i) = gamma(i) + rate(i) * weight

         endif

c          

c        ...lambda_{i+1} to lambda_{k-1}

         do m = i+1, k-1

            if (k .gt. kexact) then

               m_tilde = m + I_L2

               weight = const * power(-1, m-i-1) * 8

     &           * (   power(17, I_R1+ I_L1) / p3(m_tilde - i_tilde + 1)

     &           +  power(17, I_R1) * power(-1, I_Lge1) 

     &                 / ( p3(i_tilde) * p3(m_tilde-1) )

     &           - power(17, I_L1) * power(-1, I_Rge1) 

     &                 / ( p3(k_ring-i_tilde) * p3(k_ring-m_tilde+1) ) )

c           if (i.eq.1) print *,' m approx: m, k,weight =',m,k,weight

            else

               tmp1 =   power( 17, I_L1 )   * p3(i_tilde-1) 

     &                + power( -1, I_Lge1 ) / p3(i_tilde-1)

               tmp2 =   power( 17, I_R1 )   * p3(k - m - 1 + I_R2) 

     &                - power( -1, I_Rge1 ) / p3(k - m + 1 + I_R2)

               weight = const * eight * power(-1, m-i-1) * tmp1 * tmp2

c               print *,'  i+1--k-1 exact: weight =', weight

            endif

            gamma(i) = gamma(i) + rate(m) * weight

         enddo

c      

c        ...lambda_k = rate(k)

         if (k .gt. kexact) then

            weight = const * power(-1, k-i) * power(17, I_L1) 

     &        * ( 1 - (third + threef*I_Rge1 + fourth*I_R2) * I_k )   

     &        * ( power( 17, I_R1 ) * power( 3, I_R2 )

     &        * ( 1- (three + threef*I_Rge1 + fourth*I_R2) * (1 - I_k) )

     &        + power( -1, I_Rge1 )

     &        * ( 1 - (third + threef*I_Rge1 + fourth*I_R2) *(1 - I_k) ) 

     &            / power( 3, I_R2 )  ) / power( 3, k_ring - i_tilde )

c            if (i .eq. 1) print *,'  k approx: k, weight =', k, weight

         else

            weight = const * power( -1, k-i ) 

c

     &        * (      power( 17, I_L1 ) * p3(i_tilde-1)

     &           * ( 1 - (third + threef*I_Rge1 + fourth*I_R2) * I_k )      

     &                 + power( -1, I_Lge1 )

     &           * ( 1 - (three + threef*I_Rge1 + fourth*I_R2) * I_k ) 

     &                                / p3(i_tilde-1)                  )

c

     &        * (      power( 17, I_R1 ) * power( 3, I_R2 )

     &           * ( 1 - (three + threef*I_Rge1 + fourth*I_R2)*(1-I_k) ) 

     &                 + power( -1, I_Rge1 )

     &           * ( 1 - (third + threef*I_Rge1 + fourth*I_R2)*(1-I_k) ) 

     &                                / power( 3, I_R2 )               )       

c            print *,'  k exact: weight =', weight

         endif

         gamma(i) = gamma(i) + rate(k) * weight 

c          

c        ...lambda_{k+1} = frate(3), the inner-right fixed rate

         if (k .gt. kexact) then

            weight = const * power(-1, k-i) * power(17, I_L1) 

     &         * (threef * I_Rge1+half*I_R2) * ( power(17, I_R1)

     &         * power(3,I_R2) + power(-1, I_Rge1) / power(3,I_R2) )

     &                 / power(3,k_ring - i_tilde)

c            if (i .eq. 1) print *,'  k+1 approx: k, weight =', k, weight

         else

            weight = const * (threef*I_Rge1+half*I_R2) * power(-1, k-i) 

     &                * (   power( 17, I_L1   ) * p3(i_tilde-1) 

     &                    + power( -1, I_Lge1 ) / p3(i_tilde-1)      )

     &                * (   power( 17, I_R1   ) * power( 3, I_R2 ) 

     &                    + power( -1, I_Rge1 ) / power( 3, I_R2 )   )

c            print *,'  k+1 exact: weight =', weight

         endif

         gamma(i) = gamma(i) + frate(3) * weight

c           

c        ...lambda_{k+2} = frate(4), the outer-right fixed rate

         if (k .gt. kexact) then

            weight = const * two * I_R2 * power( -1, k-i-1 ) 

     &                 * power( 17, I_L1 ) / p3(k_ring - i_tilde+1)

c            if (i .eq. 1) print *,'  k+2 approx: k, weight =', k, weight

         else

            if (i_tilde < 2) then

               tmp1 = power( 17, I_L1 ) / p3(2 - i_tilde) 

            else

               tmp1 = power( 17, I_L1 ) * p3(i_tilde - 2) 

            endif

            tmp2 = power( -1, I_Lge1 ) / p3(i_tilde)

            weight = const * two * I_R2 * power( -1, k-i-1 )*(tmp1+tmp2)

c            print *,'  k+2 exact: weight =', weight

         endif

         gamma(i) = gamma(i) + frate(4) * weight

c

      enddo

      return

      end





      subroutine csmooth( k, rate, gamma )

c     huifen chen and bruce schmeiser, may 14, 2011.

c     purpose: compute the optimal gamma values for the cyclic 

c              context, ignoring the nonnegativity constraints.

c     input:  k     = dimension of gamma

c             rate  = vector of initial rates

c     output: gamma = optimal gamma vector 

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension rate(1), gamma(1)

      data kexact /100/

      print *,' '

      print *,'   ...beginning csmooth'

c

      if (k .le. kexact) then

c        ...exact formula

         power1 = (-3d0)**(k-2)

         power2 = 1d0

         const  = 2d0 / ( 9d0 * power1 - power2 )

      else

c        ..approximation to avoid overflow when k is large

         power1 = 1d0

         power2 = 0d0

         const  = 2d0 / 9d0

      endif

c

      do i = 1, k

         gamma(i) = 0d0

      enddo

c

      do j = 1, int( (k-1) / 2 )

         weight = const * ( power1 - power2 )

         do i = 1, k

            ipj = i + j

            imj = i - j

            if (ipj > k) ipj = ipj - k

            if (imj < 1) imj = imj + k

            gamma(i) = gamma(i) + weight * ( rate(ipj) - rate(imj) )

         enddo

         power1 = power1 / (-3d0)

         power2 = power2 * (-3d0)

      enddo

      return

      end





      function objfunc( k, rate, ifinite, frate, gamma )

c     huifen chen and bruce schmeiser, may 14, 2011.

c     purpose: compute the objective-function value

c     input:  ifinite = 1 if finite horizon, 0 if cyclic

c             k       = dimension of gamma

c             rate    = vector of initial rates

c             frate   = four fixed rates, left to right

c             gamma   = vector of rate increments

c     output: objfunc = objective-function value

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension rate(1), frate(4), gamma(1)

      objfunc = 0d0

      do i = 1, k-1

         tmp1 = rate(i) - rate(i+1) + 3d0*gamma(i) + gamma(i+1)

         tmp2 = rate(i) - rate(i+1) + gamma(i) + 3d0*gamma(i+1)

         objfunc = objfunc + tmp1**2 + tmp2**2

      enddo

c

      if (ifinite .eq. 0) then

c        ...cyclic context

         tmp1 = rate(k) - rate(1) + 3d0*gamma(k) + gamma(1)

         tmp2 = rate(k) - rate(1) + gamma(k) + 3d0*gamma(1)

         objfunc = objfunc + tmp1**2 + tmp2**2

      else

c        ...finite-horizon context (left- and right-side fixed rates)

         if (frate (2) .ge. 0d0) then

            tmp1 = frate(2) - rate(1) + 3d0*gamma(1)

            objfunc = objfunc + tmp1**2

            if (frate(1) .ge. 0d0) then

               tmp2 = -frate(1) + 2d0*frate(2) - rate(1) + gamma(1)

               objfunc = objfunc + tmp2**2

            endif

         endif

c

         if (frate (3) .ge. 0d0) then

            tmp1 = rate(k) - frate(3) + 3d0*gamma(k)

            objfunc = objfunc + tmp1**2

            if (frate(4) .ge. 0d0) then

               tmp2 = rate(k) - 2d0*frate(3) + frate(4) + gamma(k)

               objfunc = objfunc + tmp2**2

            endif

         endif

      endif

      return

      end             







      function power( i, j )

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      power = dble(i)**dble(j)

      return   

      end



      subroutine pushup1( k, rate, gamma, ipush )

c     huifen chen and bruce schmeiser, may 12, 2011.

c     purpose: push the most-negative rate up to zero.

c     input:  k, the number of time intervals.

c             rate: vector of rates

c             gamma: vector of rate shifts

c     output: gamma: updated gamma vector

c             ipush: index of the pushed gamma value, 0 if none

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension rate(1), gamma(1), ilock(1)

      print *,'   ...beginning pushup'

c

c     ...find most-negative negative new rate, if any

      ratemin = 0d0

      ipush   = 0

      do i = 1, k

         rmag = rate(i) - abs( gamma(i) )

         if (rmag .lt. ratemin) then

            ratemin = rmag

            ipush   = i

            print *,' ratemin, ipush =', ratemin, ipush

         endif

      enddo

c

c     ...push the most-negative rate, if any, to zero

      if (ipush .gt. 0)  then

         if (gamma(ipush) .lt. 0d0)  then

            gamma(ipush) = - rate(ipush)

         else

            gamma(ipush) = rate(ipush)

         endif

      endif

c      print *,'                                   ipush =', ipush

c      if (ipush .gt. 0) print *,'          gamma(ipush) =', gamma(ipush)

      return

      end





      subroutine pushup2( k, rate, ifinite, frate, gamma, ipush )

c     huifen chen and bruce schmeiser, may 14, 2011.

c     purpose: push a negative new rate up to zero.

c              choose the rate that provides the 

c              smallest objective-function value.

c     input:  k, the number of time intervals.

c             rate: vector of rates

c             ifinite: 1 if finite horizon, 0 if cyclic

c             frate: fixed rates

c             gamma: vector of rate shifts

c     output: gamma: updated gamma vector

c             ipush: index of the pushed gamma value, 0 if none

      implicit double precision (a-h,o-z)

      implicit integer (i-n)

      dimension rate(1), frate(4), gamma(1), ilock(1)

      print *,'   ...beginning pushup'

c

c     ...if any, find the potential negative new rate to push up

      ipush = 0

      do i = 1, k

c        ...check the ith interval

         if (rate(i) - abs( gamma(i) ) .lt. 0d0) then

c           ...the ith new rate would be negative

            saveg = gamma(i)

            if (gamma(i) .lt. 0d0)  then

               gamma(i) = - rate(i)

            else

               gamma(i) =   rate(i)

            endif

            obj = objfunc( k, rate, ifinite, frate, gamma )

            if (ipush .eq. 0) objmin = obj - 1d0

            if (obj .gt. objmin) then

c              ...looking for the interval with the largest obj

               objmin = obj

               bestg  = gamma(i)

               ipush  = i

            endif

            gamma(i) = saveg            

         endif

      enddo

      if (ipush .gt. 0) gamma(ipush) = bestg

      return

      end

